# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib
import os
from unittest import skipIf, skipUnless

import cherrypy
from cherrypy.test import helper
from parameterized import parameterized

import cherrypy_foundation.tools.jinja2  # noqa
from cherrypy_foundation.components import StaticMiddleware
from cherrypy_foundation.tests import SeleniumUnitTest
from cherrypy_foundation.url import url_for

HAS_JINJAX = importlib.util.find_spec("jinjax") is not None

DATATABLE_TEST = os.environ.get('DATATABLE_TEST', None)

env = cherrypy.tools.jinja2.create_env(
    package_name=__package__,
)


class Static:

    components = StaticMiddleware()


@cherrypy.tools.sessions()
@cherrypy.tools.jinja2(on=False, env=env)
class Root:

    static = Static()

    @cherrypy.expose
    def index(self, **kwargs):
        return 'OK'

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_datatable_static.html')
    def datatable_static(self):
        return {}

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_datatable_ajax.html')
    def datatable_ajax(self):
        return {}

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def data_json(self, **kwargs):
        return {
            'status': 200,
            'data': [
                {
                    "id": 1,
                    "name": "John Doe",
                    "email": "johndoe@example.com",
                    "age": 28,
                },
                {
                    "id": 2,
                    "name": "Jane Smith",
                    "email": "janesmith@example.com",
                    "age": 34,
                },
                {
                    "id": 3,
                    "name": "Bob Johnson",
                    "email": "bobjohnson@example.com",
                    "age": 45,
                },
            ],
        }


@skipUnless(HAS_JINJAX, reason='Required jinjax')
@skipIf(DATATABLE_TEST == '0', reason='Datatable test disabled')
class DatatableTest(SeleniumUnitTest, helper.CPWebCase):
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    @parameterized.expand(
        [
            ('/datatable_static',),
            ('/datatable_ajax',),
        ]
    )
    def test_get_page(self, page):
        # Given a page with datatable
        self.getPage(url_for(page))
        self.assertStatus(200)
        # Then page is render using datatable
        self.assertInBody('<table ')

    @parameterized.expand(
        [
            ('/datatable_static',),
            ('/datatable_ajax',),
        ]
    )
    def test_get_page_selenium(self, page):
        # Given  page with datatable
        with self.selenium() as driver:
            # When querying the page
            driver.get(url_for(page))
            # Then page load without error
            self.assertFalse(driver.get_log('browser'))
            # Then page contains our table
            table = driver.find_element('css selector', 'table')
            self.assertTrue(table)
            # Then the table contains our data.
            driver.find_element('xpath', "//td[text()='Bob Johnson']")


if __name__ == '__main__':
    cherrypy.quickstart(Root())



# Cherrypy-foundation
# Copyright (C) 2020-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple

from cherrypy_foundation.sessions import session_lock

FlashMessage = namedtuple('FlashMessage', ['message', 'level'])


def flash(message, level='info'):
    """
    Add a flash message to the session.
    """
    assert message
    assert level in ['info', 'error', 'warning', 'success']
    with session_lock() as session:
        if 'flash' not in session:
            session['flash'] = []
        # Support Markup and string
        if hasattr(message, '__html__'):
            flash_message = FlashMessage(message, level)
        else:
            flash_message = FlashMessage(str(message), level)
        session['flash'].append(flash_message)


def get_flashed_messages():
    """
    Return all flash message.
    """
    with session_lock() as session:
        if 'flash' in session:
            messages = session['flash']
            del session['flash']
            return messages
    return []



# OAuth Plugins for cherrypy
# Copyright (C) 2025-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from collections import namedtuple

import cherrypy
import responses
from cherrypy.test import helper

from ...tools import auth  # noqa
from .. import oauth  # noqa

UserObj = namedtuple('UserObj', 'username,email,fullname')


def user_lookup_func(login, user_info):
    if login in 'test@example.com':
        # Called with email by OAuth
        user = UserObj(username='my-user', email=user_info['email'], fullname=user_info['fullname'])
        return user.username, user
    elif login == 'my-user':
        # Called with username to restore session.
        user = UserObj(username='my-user', email='test@example.com', fullname='Test User')
        return user.username, user
    return None


def user_from_key_func(user_key):
    if user_key == 'my-user':
        # Called with username to restore session.
        user = UserObj(username='my-user', email='test@example.com', fullname='Test User')
        return user
    return None


@cherrypy.tools.auth(on=True, user_lookup_func=user_lookup_func, user_from_key_func=user_from_key_func)
@cherrypy.tools.sessions()
class Root:
    @cherrypy.expose
    def index(self):
        return "OK"


class OAuthPluginTest(helper.CPWebCase):

    interactive = False

    def setUp(self):
        # Mock OAuth provider.
        responses.add(
            responses.POST,
            "https://mock-provider.com/oauth/token",
            json={
                "access_token": "mock_access_token",
                "token_type": "Bearer",
                "expires_in": 3600,
                "refresh_token": "mock_refresh_token",
            },
            status=200,
        )
        responses.add(
            responses.GET,
            "https://mock-provider.com/user",
            json={"id": "12345", "email": "test@example.com", "name": "Test User"},
            status=200,
            headers={"Authorization": "Bearer mock_access_token"},
        )
        responses.start()
        return super().setUp()

    def tearDown(self):
        responses.stop()
        responses.reset()
        return super().tearDown()

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                # Mock OAuth
                'oauth.client_id': "test_client_id",
                'oauth.client_secret': "test_client_secret",
                'oauth.auth_url': "https://mock-provider.com/oauth/authorize",
                'oauth.token_url': "https://mock-provider.com/oauth/token",
                'oauth.userinfo_url': "https://mock-provider.com/user",
                'oauth.fullname_claim': "name",
            }
        )
        cherrypy.tree.mount(Root(), '/')

    def test_oauth_login_success(self):
        # Given a OAuth provider
        # When trying to authenticate using OAuth
        self.getPage('/oauth/login')
        # Then user get redirected to OAuth provider
        self.assertStatus(303)
        redirect = self.assertHeader('Location')
        self.assertIn(
            'https://mock-provider.com/oauth/authorize?response_type=code&client_id=test_client_id&redirect_uri=http%3A%2F%2F127.0.0.1%3A54583%2Foauth%2Fcallback&scope=openid+profile+email&state=',
            redirect,
        )
        state = redirect.split('state=')[1]
        # When OAuth provider callback
        self.getPage(f'/oauth/callback?code=test&state={state}', headers=self.cookies)
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', 'http://%s:%s/' % (self.HOST, self.PORT))
        # Then user is authenticated
        self.getPage('/', headers=self.cookies)
        self.assertStatus(200)
        self.assertBody('OK')

    def test_oauth_wrong_state(self):
        # Given a OAuth provider
        # When trying to authenticate using OAuth
        self.getPage('/oauth/login')
        # Then user get redirected to OAuth provider
        self.assertStatus(303)
        redirect = self.assertHeader('Location')
        self.assertIn(
            'https://mock-provider.com/oauth/authorize?response_type=code&client_id=test_client_id&redirect_uri=http%3A%2F%2F127.0.0.1%3A54583%2Foauth%2Fcallback&scope=openid+profile+email&state=',
            redirect,
        )
        # When OAuth provider callback with invalid state
        self.getPage('/oauth/callback?code=test&state=INVALID', headers=self.cookies)
        # Then http error 400 get raised
        self.assertStatus(400)
        self.assertInBody("Invalid state parameter")

    def test_oauth_missing_code(self):
        # Given a OAuth provider
        # When trying to authenticate using OAuth
        self.getPage('/oauth/login')
        # Then user get redirected to OAuth provider
        self.assertStatus(303)
        redirect = self.assertHeader('Location')
        self.assertIn(
            'https://mock-provider.com/oauth/authorize?response_type=code&client_id=test_client_id&redirect_uri=http%3A%2F%2F127.0.0.1%3A54583%2Foauth%2Fcallback&scope=openid+profile+email&state=',
            redirect,
        )
        # When OAuth provider callback with missing code.
        state = redirect.split('state=')[1]
        self.getPage(f'/oauth/callback?state={state}', headers=self.cookies)
        # Then http error 400 get raised
        self.assertStatus(400)
        self.assertInBody("No authorization code received")

    def test_oauth_with_error(self):
        # Given a OAuth provider
        # When trying to authenticate using OAuth
        self.getPage('/oauth/login')
        # Then user get redirected to OAuth provider
        self.assertStatus(303)
        redirect = self.assertHeader('Location')
        self.assertIn(
            'https://mock-provider.com/oauth/authorize?response_type=code&client_id=test_client_id&redirect_uri=http%3A%2F%2F127.0.0.1%3A54583%2Foauth%2Fcallback&scope=openid+profile+email&state=',
            redirect,
        )
        # When OAuth provider callback with error code.
        self.getPage('/oauth/callback?error=invalid_request', headers=self.cookies)
        # Then http error 400 get raised
        self.assertStatus(400)
        self.assertInBody("Authentication failed. OAuth error: invalid_request")



# Cherrypy-foundation
# Copyright (C) 2025-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import os
import time
from contextlib import contextmanager

import cherrypy
import zc.lockfile
from cherrypy.lib import locking
from cherrypy.lib.sessions import FileSession as CPFileSession


class FileSession(CPFileSession):
    """
    Override implementation of cherrpy file session to improve file locking.
    """

    def acquire_lock(self, path=None):
        """Acquire an exclusive lock on the currently-loaded session data."""
        # See Issue https://github.com/cherrypy/cherrypy/issues/2065

        if path is None:
            path = self._get_file_path()
        path += self.LOCK_SUFFIX
        checker = locking.LockChecker(self.id, self.lock_timeout)
        while not checker.expired():
            try:
                self.lock = zc.lockfile.LockFile(path)
            except zc.lockfile.LockError:
                # Sleep for 1ms only.
                time.sleep(0.001)
            else:
                break
        self.locked = True
        if self.debug:
            cherrypy.log('Lock acquired.', 'TOOLS.SESSIONS')

    def clean_up(self):
        """Also clean-up left over lock files."""
        # See Issue https://github.com/cherrypy/cherrypy/issues/1855

        # Clean-up session files.
        CPFileSession.clean_up(self)

        # Then clean-up any orphane lock files.
        suffix_len = len(self.LOCK_SUFFIX)
        files = os.listdir(self.storage_path)
        lock_files = [
            fname for fname in files if fname.startswith(self.SESSION_PREFIX) and fname.endswith(self.LOCK_SUFFIX)
        ]
        for fname in lock_files:
            session_file = fname[:-suffix_len]
            if session_file not in files:
                filepath = os.path.join(self.storage_path, fname)
                try:
                    os.unlink(filepath)
                except Exception as e:
                    cherrypy.log(f'Error deleting {filepath}: {e}', 'TOOLS.SESSIONS', severity=logging.WARNING)


@contextmanager
def session_lock():
    """
    Acquire session lock as required. Support re-intrant lock.
    """
    s = cherrypy.serving.session
    # If session.locking is undefined, default is implicit.
    is_explicit = getattr(s, 'locking', False) == 'explicit'
    if is_explicit and not s.locked:
        # When explicit, acquire session lock.
        s.acquire_lock()
        try:
            yield s
        finally:
            # When explicit, we want to save the session (with also release the lock.)
            if is_explicit:
                s.save()
                cherrypy.serving.request._sessionsaved = True
    else:
        yield s



# OAuth Plugins for cherrypy
# Copyright (C) 2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os

import cherrypy
from cherrypy.process.plugins import SimplePlugin
from oauthlib.oauth2 import OAuth2Error
from requests.exceptions import RequestException
from requests_oauthlib import OAuth2Session

from cherrypy_foundation.sessions import session_lock

OAUTH_STATE = '_oauth_state'

CONTEXT_OAUTH = 'OAUTH'

# Don't mind HTTPS verification done by oauthlib.
os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'


def _get_first_claim(user_info, claims):
    """
    Extract the first available claim from OpenID user info.

    Args:
        user_info (dict): OpenID user info dictionary containing claims
        claims (list): List of possible claim

    Returns:
        str or None: The value of the first matching claim found, or None if no match
    """
    if isinstance(claims, str):
        claims = [claims]
    assert isinstance(user_info, (dict, list))
    for claim in claims:
        if claim in user_info and user_info[claim] is not None:
            return user_info[claim]
    return None


class OAuthPlugin(SimplePlugin):

    client_id = None
    client_secret = None
    scope = ["openid", "profile", "email"]
    auth_url = None
    token_url = None
    userinfo_url = None

    fullname_claim = None
    firstname_claim = None
    lastname_claim = None
    email_claim = "email"
    userkey_claim = "email"
    required_claims = []

    def _verified_required_claim(self, user_info):
        """
        Verifies if the user_info contains all required claims with the expected value(s).
        Handles both scalar and list-type claims.

        Args:
            user_info (dict): The claims of the current user.

        Returns:
            bool: True if all required claims are satisfied, False otherwise.
        """
        if self.required_claims:
            for claim, expected_value in self.required_claims:
                value = user_info.get(claim, None)
                # If the value in user_info is a list
                if isinstance(value, list):
                    if expected_value not in value:
                        return False
                else:
                    if value != expected_value:
                        return False
        return True

    def start(self):
        """Plugin start - mount OAuth endpoints"""
        if self.client_id or self.client_secret:
            self.bus.log('Starting OAuth plugin')
            # Mount OAuth endpoints
            app = cherrypy.tree.apps.get('')
            if app and hasattr(app, 'root'):
                app.merge(
                    {
                        '/oauth': {
                            'tools.sessions.on': True,
                            'tools.auth.on': False,  # Don't protect OAuth endpoints
                            'tools.auth_mfa.on': False,
                        }
                    }
                )
                app.root.oauth = self

    def stop(self):
        self.bus.log('Stopping OAuth plugin')

    def graceful(self):
        """Reload of subscribers."""
        self.stop()
        self.start()

    @cherrypy.expose()
    def login(self, **kwargs):
        # Create a redirect URI.
        redirect_uri = cherrypy.url('callback')
        # Redirect to OAuth provider
        oauth = OAuth2Session(self.client_id, redirect_uri=redirect_uri, scope=self.scope)
        authorization_url, state = oauth.authorization_url(self.auth_url)
        # Need to store the "state" in session for verification
        with session_lock() as session:
            session[OAUTH_STATE] = state
        raise cherrypy.HTTPRedirect(authorization_url)

    @cherrypy.expose()
    def callback(self, code=None, state=None, error=None, error_description=None, **kwargs):

        # Check for error
        if error or error_description:
            cherrypy.log(f'OAuth error: {error} - {error_description}', severity=logging.WARNING, context=CONTEXT_OAUTH)
            raise cherrypy.HTTPError(400, "Authentication failed. OAuth error: " + error)

        # Verify state parameter
        with session_lock() as session:
            session_state = session.get(OAUTH_STATE, None)
            if not state or not session_state or state != session_state:
                cherrypy.log('OAuth state mismatch or missing', severity=logging.WARNING, context=CONTEXT_OAUTH)
                raise cherrypy.HTTPError(400, "Invalid state parameter")
            del session[OAUTH_STATE]

        # Check code value.
        if not code:
            cherrypy.log('no authorization code received', severity=logging.WARNING, context=CONTEXT_OAUTH)
            raise cherrypy.HTTPError(400, "No authorization code received")

        # Query the token.
        redirect_uri = cherrypy.url('/oauth/callback')
        oauth = OAuth2Session(self.client_id, state=session_state, redirect_uri=redirect_uri)
        try:
            # Don't need to store the token, since we are not calling OAuth afterward.
            oauth.fetch_token(self.token_url, client_secret=self.client_secret, code=code)
        except OAuth2Error as e:
            cherrypy.log(f'OAuth2 error: {e}', exc_info=1, severity=logging.WARNING, context=CONTEXT_OAUTH)
            raise cherrypy.HTTPError(400, "Authentication failed")
        except RequestException as e:
            cherrypy.log(
                f'network error during token fetch: {e}', exc_info=1, severity=logging.WARNING, context=CONTEXT_OAUTH
            )
            raise cherrypy.HTTPError(503, "Service temporarily unavailable")
        except Exception as e:
            cherrypy.log(
                f'unexpected error during token fetch: {e}', exc_info=1, severity=logging.WARNING, context=CONTEXT_OAUTH
            )
            raise cherrypy.HTTPError(500, "Internal server error")

        # Query user_info
        try:
            response = oauth.get(self.userinfo_url)
            response.raise_for_status()  # Raise exception for HTTP errors
            user_info = response.json()
        except Exception:
            cherrypy.log('fail to fetch user info', exc_info=1, severity=logging.WARNING, context=CONTEXT_OAUTH)
            raise cherrypy.HTTPError(400, "Failed to fetch user info")

        if not isinstance(user_info, dict):
            raise cherrypy.HTTPError(400, "Invalid user info format")

        cherrypy.log(f"OAuth user info: {user_info}", severity=logging.DEBUG, context=CONTEXT_OAUTH)

        # Verify required claims
        if not self._verified_required_claim(user_info):
            raise cherrypy.HTTPError(403, "Missing required claim")

        # Extract fields from the user_info.
        fullname = _get_first_claim(user_info, self.fullname_claim)
        if fullname is None:
            firstname = _get_first_claim(user_info, self.firstname_claim)
            lastname = _get_first_claim(user_info, self.lastname_claim)
            fullname = ' '.join([name for name in [firstname, lastname] if name])
        email = _get_first_claim(user_info, self.email_claim)
        userkey = _get_first_claim(user_info, self.userkey_claim)

        # Let fail if the provide doesn't has the required info.
        if not userkey or (isinstance(userkey, str) and not userkey.strip()):
            cherrypy.log(
                f'No usable user key in user info (claims tried: {self.userkey_claim})',
                severity=logging.WARNING,
                context=CONTEXT_OAUTH,
            )
            raise cherrypy.HTTPError(400, "Failed to determine user identifier from OAuth provider")

        # With OAuth with loging with email.
        userobj = cherrypy.tools.auth.login_with_result(
            login=userkey,
            user_info={
                'email': email,
                'fullname': fullname,
            },
        )
        # Login might fail if user doesn't exists.
        if userobj is None:
            cherrypy.log(f'login failed for email: {email}', severity=logging.WARNING, context=CONTEXT_OAUTH)
            raise cherrypy.HTTPError(403, "Authentication failed")

        # Redirect user
        raise cherrypy.tools.auth.redirect_to_original_url()


cherrypy.oauth = OAuthPlugin(cherrypy.engine)
cherrypy.oauth.subscribe()

cherrypy.config.namespaces['oauth'] = lambda key, value: setattr(cherrypy.oauth, key, value)



# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import tempfile

import cherrypy
from cherrypy.test import helper

from cherrypy_foundation.sessions import FileSession, session_lock


@cherrypy.tools.sessions(on=True, locking='explicit', storage_class=FileSession)
class Root:

    @cherrypy.expose
    def index(self, value='OK'):
        if value:
            with session_lock() as s:
                s['value'] = value
        return s['value']


class FileSessionTest(helper.CPWebCase):
    interactive = False

    @classmethod
    def setup_class(cls):
        cls.tempdir = tempfile.TemporaryDirectory(prefix='cherrypy-foundation-', suffix='-file-session-test')
        cls.storage_path = cls.tempdir.name
        super().setup_class()

    @classmethod
    def teardown_class(cls):
        cls.tempdir.cleanup()
        super().teardown_class()

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'tools.sessions.storage_path': cls.storage_path,
            }
        )
        cherrypy.tree.mount(Root(), '/')

    @property
    def _session_id(self):
        """Return session id from cookie."""
        if hasattr(self, 'cookies') and self.cookies:
            for unused, value in self.cookies:
                for part in value.split(';'):
                    key, unused, value = part.partition('=')
                    if key == 'session_id':
                        return value

    def test_get_page(self):
        # Given a page with session enabled
        # When the page get queried
        self.getPage("/")
        # Then a session is created with a id
        self.assertStatus(200)
        self.assertTrue(self._session_id)
        # Then this session is created on disk.
        s = FileSession(id=self._session_id, storage_path=self.storage_path)
        self.assertTrue(s._exists())
        # When session timeout and get clean-up
        s.acquire_lock()
        s.load()
        s.timeout = 0
        s.save()
        s.clean_up()
        # Then session get deleted
        self.assertFalse(s._exists())
        # Lock file also get deleted.
        self.assertFalse(os.path.exists(s._get_file_path() + FileSession.LOCK_SUFFIX))
